@@ -33,6 +33,7 @@
 #include <unique/uniqueapp.h>
 
 #include <bluetooth-client.h>
+#include <bluetooth-client-private.h>
 #include <bluetooth-chooser.h>
 #include <bluetooth-killswitch.h>
 
@@ -49,6 +50,9 @@ static gboolean show_icon_pref = TRUE;
 #define PREF_DIR		"/apps/bluetooth-manager"
 #define PREF_SHOW_ICON		PREF_DIR "/show_icon"
 
+#define KEYBOARD_PREFS		"gnome-keyboard-properties"
+#define MOUSE_PREFS		"gnome-mouse-properties"
+
 enum {
 	CONNECTED,
 	DISCONNECTED,
@@ -98,50 +102,53 @@ select_device_changed(BluetoothChooser *sel,
 
 void browse_callback(GObject *widget, gpointer user_data)
 {
-	GtkWidget *dialog, *selector, *button, *image;
-	char *bdaddr, *cmd;
-	int response_id;
+	char *address, *cmd;
 
-	dialog = gtk_dialog_new_with_buttons(_("Select Device to Browse"), NULL,
-					     GTK_DIALOG_NO_SEPARATOR,
-					     GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
-					     NULL);
-	button = gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Browse"), GTK_RESPONSE_ACCEPT);
-	image = gtk_image_new_from_icon_name (GTK_STOCK_CONNECT, GTK_ICON_SIZE_BUTTON);
-	gtk_button_set_image (GTK_BUTTON (button), image);
-	gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
-					  GTK_RESPONSE_ACCEPT, FALSE);
-	gtk_window_set_default_size(GTK_WINDOW(dialog), 480, 400);
+	address = g_strdup (g_object_get_data (widget, "address"));
+	if (address == NULL) {
+		GtkWidget *dialog, *selector;
+		int response_id;
 
-	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
-	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
+		dialog = gtk_dialog_new_with_buttons(_("Select Device to Browse"), NULL,
+						     GTK_DIALOG_NO_SEPARATOR,
+						     GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
+						     NULL);
+		gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Browse"), GTK_RESPONSE_ACCEPT);
+		gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
+						  GTK_RESPONSE_ACCEPT, FALSE);
+		gtk_window_set_default_size(GTK_WINDOW(dialog), 480, 400);
 
-	selector = bluetooth_chooser_new(_("Select device to browse"));
-	gtk_container_set_border_width(GTK_CONTAINER(selector), 5);
-	gtk_widget_show(selector);
-	g_object_set(selector,
-		     "show-search", TRUE,
-		     "show-device-category", TRUE,
-		     "show-device-type", TRUE,
-		     NULL);
-	g_signal_connect(selector, "selected-device-changed",
-			 G_CALLBACK(select_device_changed), dialog);
-	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), selector);
-	bluetooth_chooser_start_discovery (BLUETOOTH_CHOOSER (selector));
+		gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
+		gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
 
-	bdaddr = NULL;
-	response_id = gtk_dialog_run (GTK_DIALOG (dialog));
-	if (response_id == GTK_RESPONSE_ACCEPT)
-		g_object_get (G_OBJECT (selector), "device-selected", &bdaddr, NULL);
+		selector = bluetooth_chooser_new(_("Select device to browse"));
+		gtk_container_set_border_width(GTK_CONTAINER(selector), 5);
+		gtk_widget_show(selector);
+		g_object_set(selector,
+			     "show-searching", FALSE,
+			     "show-device-category", FALSE,
+			     "show-device-type", TRUE,
+			     "device-category-filter", BLUETOOTH_CATEGORY_PAIRED_OR_TRUSTED,
+			     "device-service-filter", "OBEXFileTransfer",
+			     NULL);
+		g_signal_connect(selector, "selected-device-changed",
+				 G_CALLBACK(select_device_changed), dialog);
+		gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), selector);
 
-	gtk_widget_destroy (dialog);
+		address = NULL;
+		response_id = gtk_dialog_run (GTK_DIALOG (dialog));
+		if (response_id == GTK_RESPONSE_ACCEPT)
+			g_object_get (G_OBJECT (selector), "device-selected", &address, NULL);
 
-	if (response_id != GTK_RESPONSE_ACCEPT)
-		return;
+		gtk_widget_destroy (dialog);
+
+		if (response_id != GTK_RESPONSE_ACCEPT)
+			return;
+	}
 
 	cmd = g_strdup_printf("%s --no-default-window \"obex://[%s]\"",
-			      "nautilus", bdaddr);
-	g_free (bdaddr);
+			      "nautilus", address);
+	g_free (address);
 
 	if (!g_spawn_command_line_async(cmd, NULL))
 		g_printerr("Couldn't execute command: %s\n", cmd);
@@ -151,7 +158,53 @@ void browse_callback(GObject *widget, gpointer user_data)
 
 void sendto_callback(GObject *widget, gpointer user_data)
 {
-	const char *command = "bluetooth-sendto";
+	GPtrArray *a;
+	GError *err = NULL;
+	guint i;
+	const char *address, *alias;
+
+	address = g_object_get_data (widget, "address");
+	alias = g_object_get_data (widget, "alias");
+
+	a = g_ptr_array_new ();
+	g_ptr_array_add (a, "bluetooth-sendto");
+	if (address != NULL) {
+		char *s;
+
+		s = g_strdup_printf ("--device=\"%s\"", address);
+		g_ptr_array_add (a, s);
+	}
+	if (address != NULL && alias != NULL) {
+		char *s;
+
+		s = g_strdup_printf ("--name=\"%s\"", alias);
+		g_ptr_array_add (a, s);
+	}
+	g_ptr_array_add (a, NULL);
+
+	if (g_spawn_async(NULL, (char **) a->pdata, NULL,
+			  G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &err) == FALSE) {
+		g_printerr("Couldn't execute command: %s\n", err->message);
+		g_error_free (err);
+	}
+
+	for (i = 1; a->pdata[i] != NULL; i++)
+		g_free (a->pdata[i]);
+
+	g_ptr_array_free (a, TRUE);
+}
+
+static void keyboard_callback(GObject *widget, gpointer user_data)
+{
+	const char *command = KEYBOARD_PREFS;
+
+	if (!g_spawn_command_line_async(command, NULL))
+		g_printerr("Couldn't execute command: %s\n", command);
+}
+
+static void mouse_callback(GObject *widget, gpointer user_data)
+{
+	const char *command = MOUSE_PREFS;
 
 	if (!g_spawn_command_line_async(command, NULL))
 		g_printerr("Couldn't execute command: %s\n", command);
@@ -221,11 +274,7 @@ killswitch_state_changed (BluetoothKillswitch *killswitch, KillswitchState state
 	gboolean bstate = FALSE;
 	const char *label, *status_label;
 
-	if (state == KILLSWITCH_STATE_UNKNOWN || state == KILLSWITCH_STATE_MIXED) {
-		sensitive = FALSE;
-		label = NULL;
-		status_label = N_("Bluetooth: Unknown");
-	} else if (state == KILLSWITCH_STATE_SOFT_BLOCKED) {
+	if (state == KILLSWITCH_STATE_SOFT_BLOCKED) {
 		label = N_("Turn On Bluetooth");
 		status_label = N_("Bluetooth: Off");
 		bstate = FALSE;
@@ -233,6 +282,10 @@ killswitch_state_changed (BluetoothKillswitch *killswitch, KillswitchState state
 		label = N_("Turn Off Bluetooth");
 		status_label = N_("Bluetooth: On");
 		bstate = TRUE;
+	} else if (state == KILLSWITCH_STATE_HARD_BLOCKED) {
+		sensitive = FALSE;
+		label = NULL;
+		status_label = N_("Bluetooth: Disabled");
 	} else {
 		g_assert_not_reached ();
 	}
@@ -256,10 +309,23 @@ killswitch_state_changed (BluetoothKillswitch *killswitch, KillswitchState state
 static GtkWidget *create_popupmenu(void)
 {
 	GObject *object;
+	GError *error = NULL;
 
 	xml = gtk_builder_new ();
-	if (gtk_builder_add_from_file (xml, "popup-menu.ui", NULL) == 0)
-		gtk_builder_add_from_file (xml, PKGDATADIR "/popup-menu.ui", NULL);
+	if (gtk_builder_add_from_file (xml, "popup-menu.ui", &error) == 0) {
+		if (error->domain == GTK_BUILDER_ERROR) {
+			g_warning ("Failed to load popup-menu.ui: %s", error->message);
+			g_error_free (error);
+			return NULL;
+		}
+		g_error_free (error);
+		error = NULL;
+		if (gtk_builder_add_from_file (xml, PKGDATADIR "/popup-menu.ui", &error) == 0) {
+			g_warning ("Failed to load popup-menu.ui: %s", error->message);
+			g_error_free (error);
+			return NULL;
+		}
+	}
 
 	gtk_builder_connect_signals (xml, NULL);
 
@@ -297,21 +363,19 @@ update_menu_items (void)
 	if (num_adapters_present == 0)
 		enabled = FALSE;
 	else
-		enabled = (num_adapters_present - num_adapters_powered) >= 0;
+		enabled = (num_adapters_present - num_adapters_powered) <= 0;
 
+	object = gtk_builder_get_object (xml, "adapter-action-group");
+	gtk_action_group_set_visible (GTK_ACTION_GROUP (object), enabled);
+	gtk_action_group_set_visible (devices_action_group, enabled);
 
-	object = gtk_builder_get_object (xml, "send-file");
-	gtk_action_set_sensitive (GTK_ACTION (object),
-				  enabled &&
-				  (program_available ("obex-data-server")
-				   || program_available ("obexd")));
+	if (enabled == FALSE)
+		return;
 
 	object = gtk_builder_get_object (xml, "browse-device");
 	gtk_action_set_sensitive (GTK_ACTION (object),
-				  enabled && program_available ("nautilus"));
+				  program_available ("nautilus"));
 
-	object = gtk_builder_get_object (xml, "setup-new");
-	gtk_action_set_sensitive (GTK_ACTION (object), enabled);
 }
 
 static void
@@ -373,8 +437,11 @@ set_device_status_label (const char *address, int connected)
 }
 
 static void
-connection_action_callback (GtkAction *action)
+connection_action_callback (BluetoothClient *_client,
+			    gboolean success,
+			    gpointer user_data)
 {
+	GtkAction *action = (GtkAction *) user_data;
 	const char *address;
 	int connected;
 
@@ -409,10 +476,10 @@ on_connect_activate (GtkAction *action, gpointer data)
 	path = g_object_get_data (G_OBJECT (action), "device-path");
 
 	if (connected == DISCONNECTED) {
-		if (bluetooth_client_connect_service (client, path, (BluetoothClientConnectFunc) connection_action_callback, action) != FALSE)
+		if (bluetooth_client_connect_service (client, path, connection_action_callback, action) != FALSE)
 			connected = DISCONNECTING;
 	} else if (connected == CONNECTED) {
-		if (bluetooth_client_disconnect_service (client, path, (BluetoothClientConnectFunc) connection_action_callback, action) != FALSE)
+		if (bluetooth_client_disconnect_service (client, path, connection_action_callback, action) != FALSE)
 			connected = CONNECTING;
 	} else {
 		g_assert_not_reached ();
@@ -448,9 +515,21 @@ static void
 remove_action_item (GtkAction *action, GtkUIManager *manager)
 {
 	guint menu_merge_id;
+	GList *actions, *l;
 
 	menu_merge_id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (action), "merge-id"));
 	gtk_ui_manager_remove_ui (GTK_UI_MANAGER (manager), menu_merge_id);
+	actions = gtk_action_group_list_actions (devices_action_group);
+	for (l = actions; l != NULL; l = l->next) {
+		GtkAction *a = l->data;
+		/* Don't remove the top-level action straight away */
+		if (g_str_equal (gtk_action_get_name (a), gtk_action_get_name (action)) != FALSE)
+			continue;
+		/* But remove all the sub-actions for it */
+		if (g_str_has_prefix (gtk_action_get_name (a), gtk_action_get_name (action)) != FALSE)
+			gtk_action_group_remove_action (devices_action_group, a);
+	}
+	g_list_free (actions);
 	gtk_action_group_remove_action (devices_action_group, action);
 }
 
@@ -469,10 +548,99 @@ escape_label_for_action (const char *alias)
 	return name;
 }
 
+static gboolean
+device_has_uuid (const char **uuids, const char *uuid)
+{
+	guint i;
+
+	if (uuids == NULL)
+		return FALSE;
+
+	for (i = 0; uuids[i] != NULL; i++) {
+		if (g_str_equal (uuid, uuids[i]) != FALSE)
+			return TRUE;
+	}
+	return FALSE;
+}
+
+static gboolean
+device_has_submenu (const char **uuids, GHashTable *services, BluetoothType type)
+{
+	if (services != NULL)
+		return TRUE;
+	if (device_has_uuid (uuids, "OBEXObjectPush") != FALSE)
+		return TRUE;
+	if (device_has_uuid (uuids, "OBEXFileTransfer") != FALSE)
+		return TRUE;
+	if (type == BLUETOOTH_TYPE_KEYBOARD ||
+	    type == BLUETOOTH_TYPE_MOUSE)
+	    	return TRUE;
+	return FALSE;
+}
+
+static GtkAction *
+add_menu_item (const char *address,
+	       const char *suffix,
+	       const char *label,
+	       GtkUIManager *uimanager,
+	       guint merge_id,
+	       GCallback callback)
+{
+	GtkAction *action;
+	char *action_path, *action_name;
+
+	g_return_val_if_fail (address != NULL, NULL);
+	g_return_val_if_fail (suffix != NULL, NULL);
+	g_return_val_if_fail (label != NULL, NULL);
+
+	action_name = g_strdup_printf ("%s-%s", address, suffix);
+	action = gtk_action_new (action_name, label, NULL, NULL);
+
+	gtk_action_group_add_action (devices_action_group, action);
+	g_object_unref (action);
+
+	action_path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s", address);
+	gtk_ui_manager_add_ui (uimanager, merge_id,
+			       action_path, action_name, action_name,
+			       GTK_UI_MANAGER_MENUITEM, FALSE);
+	g_free (action_path);
+
+	g_free (action_name);
+
+	if (callback != NULL)
+		g_signal_connect (G_OBJECT (action), "activate", callback, NULL);
+
+	g_object_set_data_full (G_OBJECT (action),
+				"address", g_strdup (address), g_free);
+
+	return action;
+}
+
+static void
+add_separator_item (const char *address,
+		    const char *suffix,
+		    GtkUIManager *uimanager,
+		    guint merge_id)
+{
+	char *action_path, *action_name;
+
+	g_return_if_fail (address != NULL);
+	g_return_if_fail (suffix != NULL);
+
+	action_name = g_strdup_printf ("%s-%s", address, suffix);
+	action_path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s", address);
+
+	gtk_ui_manager_add_ui (uimanager, merge_id,
+			       action_path, action_name, NULL,
+			       GTK_UI_MANAGER_SEPARATOR, FALSE);
+	g_free (action_path);
+	g_free (action_name);
+}
+
 static void
 update_device_list (GtkTreeIter *parent)
 {
-	GObject *object;
+	GtkUIManager *uimanager;
 	GtkTreeIter iter;
 	gboolean cont;
 	guint num_devices;
@@ -480,12 +648,12 @@ update_device_list (GtkTreeIter *parent)
 
 	num_devices = 0;
 
-	object = gtk_builder_get_object (xml, "bluetooth-applet-ui-manager");
+	uimanager = GTK_UI_MANAGER (gtk_builder_get_object (xml, "bluetooth-applet-ui-manager"));
 
 	if (parent == NULL) {
 		/* No default adapter? Remove everything */
 		actions = gtk_action_group_list_actions (devices_action_group);
-		g_list_foreach (actions, (GFunc) remove_action_item, object);
+		g_list_foreach (actions, (GFunc) remove_action_item, uimanager);
 		g_list_free (actions);
 		goto done;
 	}
@@ -501,158 +669,206 @@ update_device_list (GtkTreeIter *parent)
 
 	cont = gtk_tree_model_iter_children (devices_model, &iter, parent);
 	while (cont) {
-		GHashTable *table;
+		GHashTable *services;
 		DBusGProxy *proxy;
-		char *alias, *address;
+		char *alias, *address, **uuids, *name;
 		gboolean is_connected;
+		BluetoothType type;
 		GtkAction *action, *status, *oper;
 
-		action = NULL;
-
 		gtk_tree_model_get (devices_model, &iter,
 				    BLUETOOTH_COLUMN_PROXY, &proxy,
 				    BLUETOOTH_COLUMN_ADDRESS, &address,
-				    BLUETOOTH_COLUMN_SERVICES, &table,
+				    BLUETOOTH_COLUMN_SERVICES, &services,
 				    BLUETOOTH_COLUMN_ALIAS, &alias,
+				    BLUETOOTH_COLUMN_UUIDS, &uuids,
+				    BLUETOOTH_COLUMN_TYPE, &type,
 				    -1);
 
-		if (address != NULL) {
-			action = gtk_action_group_get_action (devices_action_group, address);
-			if (action) {
-				char *action_name;
+		if (device_has_submenu ((const char **) uuids, services, type) == FALSE ||
+		    address == NULL || proxy == NULL || alias == NULL) {
+			if (proxy != NULL)
+				g_object_unref (proxy);
 
-				actions = g_list_remove (actions, action);
+			if (services != NULL)
+				g_hash_table_unref (services);
+			g_strfreev (uuids);
+			g_free (alias);
+			g_free (address);
+			cont = gtk_tree_model_iter_next (devices_model, &iter);
+			continue;
+		}
 
-				action_name = g_strdup_printf ("%s-status", address);
-				status = gtk_action_group_get_action (devices_action_group, action_name);
-				g_free (action_name);
+		action = gtk_action_group_get_action (devices_action_group, address);
+		oper = NULL;
+		status = NULL;
+		if (action) {
+			char *action_name;
 
-				action_name = g_strdup_printf ("%s-action", address);
-				oper = gtk_action_group_get_action (devices_action_group, action_name);
-				g_free (action_name);
-			}
+			actions = g_list_remove (actions, action);
+
+			action_name = g_strdup_printf ("%s-status", address);
+			status = gtk_action_group_get_action (devices_action_group, action_name);
+			g_free (action_name);
+
+			action_name = g_strdup_printf ("%s-action", address);
+			oper = gtk_action_group_get_action (devices_action_group, action_name);
+			g_free (action_name);
 		}
 
 		/* If one service is connected, then we're connected */
 		is_connected = FALSE;
-		if (table != NULL) {
+		if (services != NULL) {
 			GList *list, *l;
 
-			list = g_hash_table_get_values (table);
+			list = g_hash_table_get_values (services);
 			for (l = list; l != NULL; l = l->next) {
-				gboolean val = GPOINTER_TO_INT (l->data);
-				if (val != FALSE) {
+				BluetoothStatus val = GPOINTER_TO_INT (l->data);
+				if (val == BLUETOOTH_STATUS_CONNECTED ||
+				    val == BLUETOOTH_STATUS_PLAYING) {
 					is_connected = TRUE;
 					break;
 				}
 			}
+			g_list_free (list);
 		}
 
-		if (table != NULL && address != NULL && proxy != NULL && alias != NULL) {
-			char *name;
+		name = escape_label_for_action (alias);
 
-			name = escape_label_for_action (alias);
+		if (action == NULL) {
+			guint menu_merge_id;
+			char *action_path;
 
-			if (action == NULL) {
-				guint menu_merge_id;
-				char *action_name, *action_path;
+			/* The menu item with descendants */
+			action = gtk_action_new (address, name, NULL, NULL);
 
-				/* The menu item with descendants */
-				action = gtk_action_new (address, name, NULL, NULL);
+			gtk_action_group_add_action (devices_action_group, action);
+			g_object_unref (action);
+			menu_merge_id = gtk_ui_manager_new_merge_id (uimanager);
+			gtk_ui_manager_add_ui (uimanager, menu_merge_id,
+					       "/bluetooth-applet-popup/devices-placeholder", address, address,
+					       GTK_UI_MANAGER_MENU, FALSE);
+			g_object_set_data_full (G_OBJECT (action),
+						"merge-id", GUINT_TO_POINTER (menu_merge_id), NULL);
 
-				gtk_action_group_add_action (devices_action_group, action);
-				g_object_unref (action);
-				menu_merge_id = gtk_ui_manager_new_merge_id (GTK_UI_MANAGER (object));
-				gtk_ui_manager_add_ui (GTK_UI_MANAGER (object), menu_merge_id,
-						       "/bluetooth-applet-popup/devices-placeholder", address, address,
-						       GTK_UI_MANAGER_MENU, FALSE);
-				g_object_set_data_full (G_OBJECT (action),
-							"merge-id", GUINT_TO_POINTER (menu_merge_id), NULL);
+			/* The status menu item */
+			status = add_menu_item (address,
+						"status",
+						is_connected ? _("Connected") : _("Disconnected"),
+						uimanager,
+						menu_merge_id,
+						NULL);
+			gtk_action_set_sensitive (status, FALSE);
 
-				/* The status menu item */
-				action_name = g_strdup_printf ("%s-status", address);
-				status = gtk_action_new (action_name,
-							 is_connected ? _("Connected") : _("Disconnected"), NULL, NULL);
-				gtk_action_set_sensitive (status, FALSE);
-
-				gtk_action_group_add_action (devices_action_group, status);
-				g_object_unref (status);
-
-				action_path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s", address);
-				gtk_ui_manager_add_ui (GTK_UI_MANAGER (object), menu_merge_id,
-						       action_path, action_name, action_name,
-						       GTK_UI_MANAGER_MENUITEM, FALSE);
+			if (services != NULL) {
+				action_path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s/%s-status",
+							       address, address);
+				action_set_bold (uimanager, status, action_path);
 				g_free (action_path);
-
-				action_path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s/%s",
-							       address, action_name);
-				action_set_bold (GTK_UI_MANAGER (object), status, action_path);
-				g_free (action_path);
-
-				g_free (action_name);
-
-				/* The connect button */
-				action_name = g_strdup_printf ("%s-action", address);
-				oper = gtk_action_new (action_name,
-						       is_connected ? _("Disconnect") : _("Connect"), NULL, NULL);
-
-				gtk_action_group_add_action (devices_action_group, oper);
-				g_object_unref (oper);
-
-				action_path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s", address);
-				gtk_ui_manager_add_ui (GTK_UI_MANAGER (object), menu_merge_id,
-						       action_path, action_name, action_name,
-						       GTK_UI_MANAGER_MENUITEM, FALSE);
-				g_free (action_path);
-
-				g_free (action_name);
-
-				g_signal_connect (G_OBJECT (oper), "activate",
-						  G_CALLBACK (on_connect_activate), NULL);
 			} else {
-				g_assert (oper != NULL);
-				g_assert (status != NULL);
-				gtk_action_set_label (action, name);
+				gtk_action_set_visible (status, FALSE);
+			}
+
+			/* The connect button */
+			oper = add_menu_item (address,
+					      "action",
+					      is_connected ? _("Disconnect") : _("Connect"),
+					      uimanager,
+					      menu_merge_id,
+					      G_CALLBACK (on_connect_activate));
+			if (services == NULL)
+				gtk_action_set_visible (oper, FALSE);
+
+			add_separator_item (address, "connect-sep", uimanager, menu_merge_id);
+
+			/* The Send to... button */
+			if (device_has_uuid ((const char **) uuids, "OBEXObjectPush") != FALSE) {
+				add_menu_item (address,
+					       "sendto",
+					       _("Send files..."),
+					       uimanager,
+					       menu_merge_id,
+					       G_CALLBACK (sendto_callback));
+				g_object_set_data_full (G_OBJECT (action),
+							"alias", g_strdup (alias), g_free);
+			}
+			if (device_has_uuid ((const char **) uuids, "OBEXFileTransfer") != FALSE) {
+				add_menu_item (address,
+					       "browse",
+					       _("Browse files..."),
+					       uimanager,
+					       menu_merge_id,
+					       G_CALLBACK (browse_callback));
+			}
+
+			add_separator_item (address, "files-sep", uimanager, menu_merge_id);
+
+			if (type == BLUETOOTH_TYPE_KEYBOARD && program_available (KEYBOARD_PREFS)) {
+				add_menu_item (address,
+					       "keyboard",
+					       _("Open Keyboard Preferences..."),
+					       uimanager,
+					       menu_merge_id,
+					       G_CALLBACK (keyboard_callback));
+			}
+			if (type == BLUETOOTH_TYPE_MOUSE && program_available (MOUSE_PREFS)) {
+				add_menu_item (address,
+					       "mouse",
+					       _("Open Mouse Preferences..."),
+					       uimanager,
+					       menu_merge_id,
+					       G_CALLBACK (mouse_callback));
+			}
+		} else {
+			gtk_action_set_label (action, name);
+
+			gtk_action_set_visible (status, services != NULL);
+			gtk_action_set_visible (oper, services != NULL);
+			if (services != NULL) {
 				set_device_status_label (address, is_connected ? CONNECTED : DISCONNECTED);
 				gtk_action_set_label (oper, is_connected ? _("Disconnect") : _("Connect"));
 			}
-			g_free (name);
+		}
+		g_free (name);
 
+		if (oper != NULL) {
 			g_object_set_data_full (G_OBJECT (oper),
 						"connected", GINT_TO_POINTER (is_connected ? CONNECTED : DISCONNECTED), NULL);
-			g_object_set_data_full (G_OBJECT (oper),
-						"address", g_strdup (address), g_free);
 			g_object_set_data_full (G_OBJECT (oper),
 						"device-path", g_strdup (dbus_g_proxy_get_path (proxy)), g_free);
-
-			/* And now for the trick of the day */
-			if (is_connected != FALSE) {
-				char *path;
-
-				path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s", address);
-				action_set_bold (GTK_UI_MANAGER (object), action, path);
-				g_free (path);
-			}
-
-			num_devices++;
 		}
+
+		/* And now for the trick of the day */
+		if (is_connected != FALSE) {
+			char *path;
+
+			path = g_strdup_printf ("/bluetooth-applet-popup/devices-placeholder/%s", address);
+			action_set_bold (uimanager, action, path);
+			g_free (path);
+		}
+
+		num_devices++;
+
 		if (proxy != NULL)
 			g_object_unref (proxy);
 
+		if (services != NULL)
+			g_hash_table_unref (services);
+		g_strfreev (uuids);
 		g_free (alias);
 		g_free (address);
 		cont = gtk_tree_model_iter_next (devices_model, &iter);
 	}
 
 	/* Remove the left-over devices */
-	g_list_foreach (actions, (GFunc) remove_action_item, object);
+	g_list_foreach (actions, (GFunc) remove_action_item, uimanager);
 	g_list_free (actions);
 
 done:
-	gtk_ui_manager_ensure_update (GTK_UI_MANAGER (object));
+	gtk_ui_manager_ensure_update (uimanager);
 
-	object = gtk_builder_get_object (xml, "devices-label");
-	gtk_action_set_visible (GTK_ACTION (object), num_devices > 0);
+	gtk_action_set_visible (GTK_ACTION (gtk_builder_get_object (xml, "devices-label")),
+				num_devices > 0);
 }
 
 static void device_changed (GtkTreeModel *model,
@@ -792,6 +1008,10 @@ int main(int argc, char *argv[])
 			  G_CALLBACK (device_changed), NULL);
 	/* Set the default */
 	device_changed (devices_model, NULL, NULL, NULL);
+	if (killswitch != NULL) {
+		killswitch_state_changed (killswitch,
+					  bluetooth_killswitch_get_state (killswitch));
+	}
 
 	gconf = gconf_client_get_default();
 
